@@ -1,8 +1,9 @@
 """Main module."""
 import asyncio
-from datetime import datetime, timedelta, timezone
 import json
 import logging
+from datetime import datetime, timedelta, timezone
+from http import HTTPStatus
 from typing import Any, Dict, List, Optional, Union
 
 from aiohttp import ClientConnectionError, ClientSession
@@ -11,13 +12,17 @@ from .const import (
     BASE_URL_V4,
     CURRENT,
     DAILY,
-    FIELDS_V4,
     FORECASTS,
+    HEADER_DAILY_API_LIMIT,
     HEADERS,
     HOURLY,
     NOWCAST,
+    ONE_DAY,
+    ONE_HOUR,
+    REALTIME,
     TIMESTEP_DAILY,
     TIMESTEP_HOURLY,
+    VALID_TIMESTEPS,
 )
 from .exceptions import (
     CantConnectException,
@@ -27,57 +32,31 @@ from .exceptions import (
     RateLimitedException,
     UnknownException,
 )
+from .fields import FIELDS_V4
 from .helpers import async_to_sync
 
 _LOGGER = logging.getLogger(__name__)
 
 
-def process_v4_fields(fields: List[str], timestep: str) -> str:
+def process_v4_fields(
+    fields: List[str], timestep: timedelta, write_log: bool = True
+) -> List[str]:
     """
     Filter v4 field list to only include valid fields for a given endpoint.
 
-    Logs a warning when fields get filtered out.
+    Optionally logs a warning when fields get filtered out.
     """
     valid_fields = [field for field in fields if field in FIELDS_V4]
-    if len(valid_fields) < len(fields):
+    if write_log and len(valid_fields) < len(fields):
         _LOGGER.warning(
             "Removed invalid fields: %s", list(set(fields) - set(valid_fields))
         )
-
-    if timestep == timedelta(days=1):
-        processed_fields = [
-            field for field in valid_fields if FIELDS_V4[field]["timestep"][1] == 360
-        ]
-    elif timestep == timedelta(hours=1):
-        processed_fields = [
-            field for field in valid_fields if FIELDS_V4[field]["timestep"][1] >= 108
-        ]
-    elif timestep in (
-        timedelta(minutes=30),
-        timedelta(minutes=15),
-        timedelta(minutes=5),
-        timedelta(minutes=1),
-    ):
-        processed_fields = [
-            field for field in valid_fields if FIELDS_V4[field]["timestep"][1] >= 6
-        ]
-    elif timestep == timedelta(0):
-        processed_fields = [
-            field
-            for field in valid_fields
-            if FIELDS_V4[field]["timestep"][0] <= 0
-            and FIELDS_V4[field]["timestep"][1] >= 0
-        ]
-    elif timestep < timedelta(0):
-        processed_fields = [
-            field for field in valid_fields if FIELDS_V4[field]["timestep"][0] < 0
-        ]
-    else:
-        raise InvalidTimestep
-
-    if len(processed_fields) < len(valid_fields):
+    processed_fields = [
+        field for field in valid_fields if timestep <= FIELDS_V4[field].max_timestep
+    ]
+    if write_log and len(processed_fields) < len(valid_fields):
         _LOGGER.warning(
-            "Remove fields not available for `%s` timestep: %s",
+            "Removed fields not available for `%s` timestep: %s",
             timestep,
             list(set(valid_fields) - set(processed_fields)),
         )
@@ -100,32 +79,37 @@ class TomorrowioV4:
         latitude: Union[int, float, str],
         longitude: Union[int, float, str],
         unit_system: str = "imperial",
-        session: ClientSession = None,
+        session: Optional[ClientSession] = None,
     ) -> None:
         """Initialize Tomorrow.io API object."""
         if unit_system.lower() not in ("metric", "imperial", "si", "us"):
             raise ValueError("`unit_system` must be `metric` or `imperial`")
-        elif unit_system.lower() == "si":
+        if unit_system.lower() == "si":
             unit_system = "metric"
         elif unit_system.lower() == "us":
             unit_system = "imperial"
 
         self._apikey = apikey
-        self.location = [float(latitude), float(longitude)]
         self.unit_system = unit_system.lower()
         self._session = session
         self._params = {
-            "location": self.location,
+            "location": f"{float(latitude)},{float(longitude)}",
             "units": self.unit_system,
         }
         self._headers = {**HEADERS, "apikey": self._apikey}
+        self._max_requests_per_day: Optional[int] = None
+
+    @property
+    def max_requests_per_day(self) -> Optional[int]:
+        """Return the maximum number of requests per day."""
+        return self._max_requests_per_day
 
     @staticmethod
     def convert_fields_to_measurements(fields: List[str]) -> List[str]:
         """Converts general field list into fields with measurements."""
         field_list = []
         for field in fields:
-            measurements = FIELDS_V4[field]["measurements"]
+            measurements = FIELDS_V4[field].measurements
             if len(measurements) < 2:
                 field_list.append(field)
             else:
@@ -137,90 +121,57 @@ class TomorrowioV4:
 
     @staticmethod
     def available_fields(
-        timestep: timedelta, types: Optional[List] = None
+        timestep: timedelta, types: Optional[List[str]] = None
     ) -> List[str]:
         "Return available fields for a given timestep."
-        if timestep == timedelta(days=1):
-            fields = [
-                field for field in FIELDS_V4 if FIELDS_V4[field]["timestep"][1] == 360
-            ]
-        elif timestep == timedelta(hours=1):
-            fields = [
-                field for field in FIELDS_V4 if FIELDS_V4[field]["timestep"][1] >= 108
-            ]
-        elif timestep in (
-            timedelta(minutes=30),
-            timedelta(minutes=15),
-            timedelta(minutes=5),
-            timedelta(minutes=1),
-        ):
-            fields = [
-                field for field in FIELDS_V4 if FIELDS_V4[field]["timestep"][1] >= 6
-            ]
-        elif timestep in (timedelta(0), CURRENT):
-            fields = [
-                field
-                for field in FIELDS_V4
-                if FIELDS_V4[field][0] <= 0 and FIELDS_V4[field]["timestep"][1] >= 0
-            ]
-        elif timestep < timedelta(0):
-            fields = [
-                field for field in FIELDS_V4 if FIELDS_V4[field]["timestep"][0] < 0
-            ]
-        else:
-            raise InvalidTimestep
+        if timestep not in VALID_TIMESTEPS:
+            raise InvalidTimestep(f"{timestep} is not a valid 'timestep' parameter")
+        fields = process_v4_fields(list(FIELDS_V4), timestep, write_log=False)
 
         if types:
-            return [field for field in fields if FIELDS_V4[field]["type"] in types]
+            return [field for field in fields if FIELDS_V4[field].type in types]
 
         return fields
 
     async def _call_api(self, params: Dict[str, Any]) -> Dict[str, Any]:
         """Call Tomorrow.io API."""
-        try:
-            if self._session:
-                resp = await self._session.post(
-                    BASE_URL_V4,
-                    headers=self._headers,
-                    data=json.dumps({**self._params, **params}),
-                )
-                resp_json = await resp.json()
-                if resp.status == 200:
-                    return resp_json
-                if resp.status == 400:
-                    raise MalformedRequestException(resp_json, resp.headers)
-                elif resp.status in (401, 403):
-                    raise InvalidAPIKeyException(resp_json, resp.headers)
-                elif resp.status == 429:
-                    raise RateLimitedException(resp_json, resp.headers)
-                else:
-                    raise UnknownException(resp_json, resp.headers)
+        if self._session:
+            return await self._make_call(params, self._session)
 
-            async with ClientSession() as session:
-                resp = await session.post(
-                    BASE_URL_V4,
-                    headers=self._headers,
-                    data=json.dumps({**self._params, **params}),
-                )
-                resp_json = await resp.json()
-                if resp.status == 200:
-                    return resp_json
-                if resp.status == 400:
-                    raise MalformedRequestException(resp_json, resp.headers)
-                elif resp.status in (401, 403):
-                    raise InvalidAPIKeyException(resp_json, resp.headers)
-                elif resp.status == 429:
-                    raise RateLimitedException(resp_json, resp.headers)
-                else:
-                    raise UnknownException(resp_json, resp.headers)
-        except ClientConnectionError:
-            raise CantConnectException()
+        async with ClientSession() as session:
+            return await self._make_call(params, session)
+
+    async def _make_call(
+        self, params: Dict[str, Any], session: ClientSession
+    ) -> Dict[str, Any]:
+        try:
+            resp = await session.post(
+                BASE_URL_V4,
+                headers=self._headers,
+                data=json.dumps({**self._params, **params}),
+            )
+            resp_json = await resp.json()
+        except ClientConnectionError as error:
+            raise CantConnectException() from error
+
+        if resp.status == HTTPStatus.OK:
+            max_requests = resp.headers[HEADER_DAILY_API_LIMIT]
+            if max_requests != self._max_requests_per_day:
+                self._max_requests_per_day = int(max_requests)
+            return resp_json
+        if resp.status == HTTPStatus.BAD_REQUEST:
+            raise MalformedRequestException(resp_json, resp.headers)
+        if resp.status in (HTTPStatus.UNAUTHORIZED, HTTPStatus.FORBIDDEN):
+            raise InvalidAPIKeyException(resp_json, resp.headers)
+        if resp.status == HTTPStatus.TOO_MANY_REQUESTS:
+            raise RateLimitedException(resp_json, resp.headers)
+        raise UnknownException(resp_json, resp.headers)
 
     async def realtime(self, fields: List[str]) -> Dict[str, Any]:
         """Return realtime weather conditions from Tomorrow.io API."""
         return await self._call_api(
             {
-                "fields": process_v4_fields(fields, timedelta(0)),
+                "fields": process_v4_fields(fields, REALTIME),
                 "timesteps": ["current"],
             }
         )
@@ -232,37 +183,34 @@ class TomorrowioV4:
         start_time: Optional[datetime] = None,
         duration: Optional[timedelta] = None,
         **kwargs,
-    ) -> List[Dict[str, Any]]:
+    ) -> Dict[str, Any]:
         """Return forecast data from Tomorrow.io's API for a given time period."""
-        params = {
-            "fields": self.convert_fields_to_measurements(
-                process_v4_fields(fields, timestep)
-            ),
+        if timestep not in VALID_TIMESTEPS:
+            raise InvalidTimestep(f"{timestep} is not a valid 'timestep' parameter")
+        fields = process_v4_fields(fields, timestep)
+        if timestep > ONE_HOUR:
+            fields = self.convert_fields_to_measurements(fields)
+
+        params: Dict[str, Any] = {
+            "fields": fields,
             **kwargs,
         }
-        if timestep == timedelta(days=1):
-            params["timestep"] = [TIMESTEP_DAILY]
-        elif timestep == timedelta(hours=1):
-            params["timestep"] = [TIMESTEP_HOURLY]
-        elif timestep in (
-            timedelta(minutes=30),
-            timedelta(minutes=15),
-            timedelta(minutes=5),
-            timedelta(minutes=1),
-        ):
-            params["timestep"] = [f"{int(timestep.total_seconds()/60)}m"]
+        if timestep == ONE_DAY:
+            params["timesteps"] = [TIMESTEP_DAILY]
+        elif timestep == ONE_HOUR:
+            params["timesteps"] = [TIMESTEP_HOURLY]
         else:
-            raise InvalidTimestep
+            params["timesteps"] = [f"{int(timestep.total_seconds()/60)}m"]
 
         if start_time:
             if not start_time.tzinfo:
                 start_time.replace(tzinfo=timezone.utc)
-            params["startTime"] = f"{start_time.replace(microsecond=0).isoformat()}"
+            params["startTime"] = start_time.replace(microsecond=0).isoformat()
         else:
-            start_time = datetime.utcnow().replace(tzinfo=timezone.utc)
+            start_time = datetime.now(tz=timezone.utc)
         if duration:
             end_time = (start_time + duration).replace(microsecond=0)
-            params["endTime"] = f"{end_time.isoformat()}"
+            params["endTime"] = end_time.isoformat()
 
         return await self._call_api(params)
 
@@ -291,7 +239,7 @@ class TomorrowioV4:
     ) -> Dict[str, Any]:
         """Return daily forecast data from Tomorrow.io's API for a given time period."""
         return await self._forecast(
-            timedelta(days=1), fields, start_time=start_time, duration=duration
+            ONE_DAY, fields, start_time=start_time, duration=duration
         )
 
     async def forecast_hourly(
@@ -302,7 +250,7 @@ class TomorrowioV4:
     ) -> Dict[str, Any]:
         """Return hourly forecast data from Tomorrow.io's API for a given time period."""
         return await self._forecast(
-            timedelta(hours=1), fields, start_time=start_time, duration=duration
+            ONE_HOUR, fields, start_time=start_time, duration=duration
         )
 
     async def realtime_and_all_forecasts(
@@ -336,6 +284,7 @@ class TomorrowioV4:
             ret_data[CURRENT] = data["data"]["timelines"][0]["intervals"][0]["values"]
 
         forecasts = ret_data.setdefault(FORECASTS, {})
+        start_time = datetime.now(tz=timezone.utc).isoformat()
 
         if not hourly_fields and not daily_fields:
             data = await self._call_api(
@@ -346,9 +295,7 @@ class TomorrowioV4:
                         TIMESTEP_DAILY,
                     ],
                     "fields": forecast_or_nowcast_fields,
-                    "startTime": datetime.utcnow()
-                    .replace(tzinfo=timezone.utc)
-                    .isoformat(),
+                    "startTime": start_time,
                 }
             )
             if "data" in data and "timelines" in data["data"]:
@@ -365,9 +312,7 @@ class TomorrowioV4:
                 {
                     "timesteps": [f"{nowcast_timestep}m"],
                     "fields": forecast_or_nowcast_fields,
-                    "startTime": datetime.utcnow()
-                    .replace(tzinfo=timezone.utc)
-                    .isoformat(),
+                    "startTime": start_time,
                 }
             )
             if "data" in data and "timelines" in data["data"]:
@@ -383,9 +328,7 @@ class TomorrowioV4:
                         {
                             "timesteps": [timestep],
                             "fields": field_list,
-                            "startTime": datetime.utcnow()
-                            .replace(tzinfo=timezone.utc)
-                            .isoformat(),
+                            "startTime": start_time,
                         }
                     )
                     if "data" in data and "timelines" in data["data"]:
@@ -418,7 +361,7 @@ class TomorrowioV4Sync(TomorrowioV4):
         start_time: Optional[datetime] = None,
         duration: Optional[timedelta] = None,
         timestep: int = 5,
-    ) -> List[Dict[str, Any]]:
+    ) -> Dict[str, Any]:
         """Return forecast data from Tomorrow.io's NowCast API for a given time period."""
         return await super().forecast_nowcast(fields, start_time, duration, timestep)
 
@@ -428,7 +371,7 @@ class TomorrowioV4Sync(TomorrowioV4):
         fields: List[str],
         start_time: Optional[datetime] = None,
         duration: Optional[timedelta] = None,
-    ) -> List[Dict[str, Any]]:
+    ) -> Dict[str, Any]:
         """Return daily forecast data from Tomorrow.io's API for a given time period."""
         return await super().forecast_daily(fields, start_time, duration)
 
@@ -438,7 +381,7 @@ class TomorrowioV4Sync(TomorrowioV4):
         fields: List[str],
         start_time: Optional[datetime] = None,
         duration: Optional[timedelta] = None,
-    ) -> List[Dict[str, Any]]:
+    ) -> Dict[str, Any]:
         """Return hourly forecast data from Tomorrow.io's API for a given time period."""
         return await super().forecast_hourly(fields, start_time, duration)
 
@@ -446,7 +389,7 @@ class TomorrowioV4Sync(TomorrowioV4):
     async def realtime_and_all_forecasts(
         self,
         realtime_fields: List[str],
-        forecast_fields: List[str],
+        forecast_or_nowcast_fields: List[str],
         hourly_fields: List[str] = None,
         daily_fields: List[str] = None,
         nowcast_timestep: int = 5,
@@ -459,8 +402,8 @@ class TomorrowioV4Sync(TomorrowioV4):
         data.
         """
         return await super().realtime_and_all_forecasts(
-            realtime_fields,
-            forecast_fields,
+            realtime_fields=realtime_fields,
+            forecast_or_nowcast_fields=forecast_or_nowcast_fields,
             hourly_fields=hourly_fields,
             daily_fields=daily_fields,
             nowcast_timestep=nowcast_timestep,
